/*******************************************************************
  XY Stepper Motor HPGL Plotter

  mstep.cpp	- implementation of fast fixed-point arithmetics

  (C) Copyright 1995,1997 by Pavel Pisa - Original Author
      e-mail:	pisa@cmp.felk.cvut.cz
      homepage:	http://cmp.felk.cvut.cz/~pisa
      work:	http://www.pikron.com/

  This package can be copied and used under one of next licenses

  Gnu General Public License with all its conditions.
  See file GPL for details. Under this license nobody can
  distribute this work and any derived work without full source code
  for all libraries and modules compiled and linked into executable.
  In this case you have right to add your copyright and
  delete next paragraph from this header if you disagree with its
  conditions.
  
  There is second possibility which can be useful for embedded
  and commercial applications. You can use this code as you like
  without any warranties and distribute executables or appliances
  without source. You must send all changes and enhancements made
  in this file and all files included (only) in this module to
  original author (Pavel Pisa) without any further restrictions.

 *******************************************************************/

#define MSTEPU 10000
#define MSTEPH 100

class mstep {
  public:
  long l;
  mstep();
  mstep(long x);
  mstep(long x,long d);
  mstep operator =(mstep x);
  mstep operator =(long x);
  mstep & operator +=(mstep x);
  mstep & operator -=(mstep x);
  mstep & operator *=(mstep x);
  mstep & operator /=(mstep x);
  void disp();
 };

#include <stdlib.h>
#include <stdio.h>

template <class T>
T max(T x, T y)
{
  return (x > y) ? x : y;
};

long muldiv(long m1,long m2,long d)
{
  asm{
	mov	eax,m1
	imul	dword ptr m2
	idiv	dword ptr d
  };
  return(_EAX);
};

mstep::mstep(){l=0;};
mstep::mstep(long x){l=x*MSTEPU;};
mstep::mstep(long x,long d){l=muldiv(x,MSTEPU,d);};
//mstep mstep::operator =(mstep x){l=x.l;return(*this);};
//mstep mstep::operator =(long x){l=x*MSTEPU;return(*this);};
mstep & mstep::operator +=(mstep x){l+=x.l;return(*this);};
mstep & mstep::operator -=(mstep x){l-=x.l;return(*this);};

mstep & mstep::operator *=(mstep x)
 {
  l=muldiv(l,x.l,MSTEPU);
  return(*this);
 };

mstep & mstep::operator /=(mstep x)
 {
  l=muldiv(l,MSTEPU,x.l);
  return(*this);
 };

//mstep operator +(mstep x,mstep y){mstep z=x;z+=y;return(z);};
mstep operator +(mstep x,mstep y)
{long z;
 asm{
	mov	eax,x
	add	eax,y
	mov     z,eax
 };
 return(*(mstep*)&z);
};

mstep operator -(mstep x,mstep y)
{long z;
 asm{
	mov	eax,x
	sub	eax,y
	mov     z,eax
 };
 return(*(mstep*)&z);
};

mstep operator *(mstep x,mstep y)
{long z;
 z=muldiv(*(long*)&x,*(long*)&y,MSTEPU);
 return(*(mstep*)&z);
};

mstep operator /(mstep x,mstep y)
{long z;
 z=muldiv(*(long*)&x,MSTEPU,*(long*)&y);
 return(*(mstep*)&z);
};

mstep operator -(mstep x){mstep z;z.l=-x.l;return(z);};

long  round(mstep x){return(x.l/MSTEPU);};

mstep sinx(mstep x,int a)
{
 if ((a>=360)||(a<0)) a%=360;
 if (a>90)  a=180-a;
 if (a<-90) a=-180-a;
 if (a==0) return(0);
 if (a==90) return(x);
 if (a==-90) return(-x);
 long a1,a2;
 a1=((long)a*MSTEPU)/36;
 a2=((((long)a*MSTEPU)/36)*(long)a)/36;
 mstep z=0;
 long  k[4]={-718887,814312,-413397,62832};
 int i;
 for (i=0;i<4;i++)
 {
   z.l=muldiv(z.l,a2,MSTEPU*100L)+k[i];
 };
 z.l=muldiv(z.l+2,a1,MSTEPU*10L);
 return(z*x);
};

mstep cosx(mstep x,int a)
{
 return(sinx(x,90-a));
};

mstep vectabs(mstep x,mstep y)
{
 x.l=labs(x.l);
 y.l=labs(y.l);
 if (!x.l) return(y);
 if (!y.l) return(x);
 mstep z,w;
 z.l=max(x.l,y.l);
 w.l=0;
 int I=0;
 while((w.l!=z.l)&&(I++<10))
 {w.l=z.l;
  z.l=muldiv(x.l+2,x.l,2*w.l)+muldiv(y.l+2,y.l,2*w.l)+w.l/2;I++;
 };
 return(z);
};

mstep abs(mstep x){if(x.l<0)return(x.l=-x.l,x);else return(x);};

void mstep::disp(){printf(" %c%li.%04li",l<0?'-':' ',labs(l/MSTEPU),labs(l%MSTEPU));};

/*main()
{
 mstep x(2),y(-123,100),z,t=2;
 printf("\n\n");
 x.disp();
 y.disp();
 z.disp();
 t.disp();
 printf("\n");
 z+=y;z.disp();
 z=y+x;z.disp();
 z=y*x;z.disp();
 z=y/x;z.disp();
};*/