/*******************************************************************
  XY Stepper Motor HPGL Plotter

  xy.cpp	- main module, HPGL and BGI font interpretter

  (C) Copyright 1995,1997 by Pavel Pisa - Original Author
      e-mail:	pisa@cmp.felk.cvut.cz
      homepage:	http://cmp.felk.cvut.cz/~pisa
      work:	http://www.pikron.com/

  This package can be copied and used under one of next licenses

  Gnu General Public License with all its conditions.
  See file GPL for details. Under this license nobody can
  distribute this work and any derived work without full source code
  for all libraries and modules compiled and linked into executable.
  In this case you have right to add your copyright and
  delete next paragraph from this header if you disagree with its
  conditions.
  
  There is second possibility which can be useful for embedded
  and commercial applications. You can use this code as you like
  without any warranties and distribute executables or appliances
  without source. You must send all changes and enhancements made
  in this file and all files included (only) in this module to
  original author (Pavel Pisa) without any further restrictions.

 *******************************************************************/

#include <dos.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <signal.h>
#include <values.h>
#include <ctype.h>
#include <string.h>
#include <graphics.h>
#include <dir.h>
#include "xy_plt.h"
#include "font.h"
#include "mstep.h"

typedef unsigned int word;
typedef unsigned char byte;

template <class T>
T max(T x, T y)
{
  return (x > y) ? x : y;
};

typedef void (*PlotterT)(int X,int Y,int Pen);

void BgiMoveAbs(int X,int Y,int Pen);

void NulMoveAbs(int ,int ,int ){};

typedef struct point{
  public:
  mstep x,y;
  inline point(){x=0;y=0;};
  inline point(mstep ax,mstep ay){x=ax;y=ay;};
  inline point operator +=(point b){x+=b.x;y+=b.y;return(*this);};
  inline point operator -=(point b){x-=b.x;y-=b.y;return(*this);};
 };

point operator +(point a,point b){a+=b;return(a);}
point operator -(point a,point b){a-=b;return(a);}

typedef struct tran{
  public:
  tran  *next;
  mstep xx,xy,yx,yy,xo,yo;
  point act;
  tran();
  void DoT(point &o,point &i);
 };

tran::tran(){xx=1;xy=0;yx=0;yy=1;xo=0;yo=0;next=NULL;};

/*void tran::DoT(point &o,point &i){
 act=i;
 o.x=act.x*xx+act.y*xy+xo;
 o.y=act.x*yx+act.y*yy+yo;
};*/

void tran::DoT(point &o,point &i){
 act.x=i.x;
 act.y=i.y;

 _ECX=xx.l;
 _EAX=act.x.l;
 asm imul  ecx
 asm mov   ecx,MSTEPU
 asm idiv  ecx
 asm push  eax
 _ECX=xy.l;
 _EAX=act.y.l;
 asm imul  ecx
 asm mov   ecx,MSTEPU
 asm idiv  ecx
 asm pop   edx
 asm add   edx,eax
 _EAX=xo.l;
 asm add   edx,eax
 o.x.l=_EDX;

 _ECX=yx.l;
 _EAX=act.x.l;
 asm imul  ecx
 asm mov   ecx,MSTEPU
 asm idiv  ecx
 asm push  eax
 _ECX=yy.l;
 _EAX=act.y.l;
 asm imul  ecx
 asm mov   ecx,MSTEPU
 asm idiv  ecx
 asm pop   edx
 asm add   edx,eax
 _EAX=yo.l;
 asm add   edx,eax
 o.y.l=_EDX;
};

/*void tran::DoT(point &o,point &i){
 act=i;
 o.x.l=muldiv(act.x.l,xx.l,MSTEPU)+muldiv(act.y.l,xy.l,MSTEPU)+xo.l;
 o.y.l=muldiv(act.x.l,yx.l,MSTEPU)+muldiv(act.y.l,yy.l,MSTEPU)+yo.l;
};*/

tran *T=NULL;

void AddT(tran *&T,tran *T1){
 T1->next=T;T=T1;
};

void DelT(tran *&T){
 tran *T1=T->next;
 T->next=NULL;
 delete(T);
 T=T1;
};

void DelAllT(){
 while(T)DelT(T);
};

void DoT(tran *U,point &o,point &i){
 tran *V;
 if (U==NULL) {o=i;return;}
 U->act=i;
 while (1)
 {
  if ((V=U->next)!=NULL) U->DoT(V->act,U->act);
  else {U->DoT(o,U->act);return;}
  U=V;
 };
};

tran *TShift(mstep x,mstep y){
 tran *T=new tran;
 T->xo=x;
 T->yo=y;
 return(T);
};

tran *TRot(mstep x,mstep y,int angle,mstep qx,mstep qy){
 tran *T=new tran;
 T->xo=x;
 T->yo=y;
 T->xx=cosx(qx,angle);T->xy=sinx(qy,angle);
 T->yx=sinx(qx,angle);T->yy=cosx(qy,angle);
 return(T);
};

tran *TRot(mstep x,mstep y,mstep ax,mstep ay,mstep qx,mstep qy){
 tran *T=new tran;
 mstep a=vectabs(ax,ay);
 T->xo=x;
 T->yo=y;
 T->xx.l=muldiv(ax.l,qx.l,a.l);T->xy.l=muldiv(-ay.l,qy.l,a.l);
 T->yx.l=muldiv(ay.l,qx.l,a.l);T->yy.l=muldiv(ax.l,qy.l,a.l);
 return(T);
};

tran *TRot(mstep x,mstep y,mstep ax,mstep ay,mstep qx,mstep qy,mstep sl){
 tran *T=new tran;
 mstep a=vectabs(ax,ay);
 T->xo=x;
 T->yo=y;
 T->xx.l=muldiv(ax.l,qx.l,a.l);
 T->xy.l=muldiv(-ay.l,qy.l,a.l)+muldiv(ax.l,(qx*sl).l,a.l);
 T->yx.l=muldiv(ay.l,qx.l,a.l);
 T->yy.l=muldiv(ax.l,qy.l,a.l)+muldiv(ay.l,(qx*sl).l,a.l);
 return(T);
};

//    rozm. HP norma Ve skutecnosti
// x: 264mm=10612hpu=2640
// y: 192mm= 7721hpu=1920

point P1(  561,  306);//(  306, 181);
point P2(15761,10306);//(10306,7381);
point SI;
point DI(1,0);
mstep SL=0;
byte  ETXch=3;
int   SClev=0;
point SC1,SC2;
point CHsp(mstep(15,10),2);
mstep CHsz=1;
mstep CHq=1;
char  FontName[MAXPATH]="litt.chr";
int   Prop_Fl=0;
point SCu(1,1);
char  *ExeName=NULL;
int   PS=3;

void TP1P2(){
 DelAllT();
 tran *T1=new tran;
 if (PS<=3)
 {
//  T1->xx=mstep(2640,16158)*SCu.x;
//  T1->yy=mstep(1920,10612)*SCu.y;
  T1->xx=mstep(2640,15007)*SCu.x;
  T1->yy=mstep(1920,10919)*SCu.y;
 }else{
  T1->xx=mstep(2640,10612)*SCu.x;
  T1->yy=mstep(1920, 7721)*SCu.y;
 };
 T1->xo=T1->xx*P1.x;
 T1->yo=T1->yy*P1.y;
 AddT(T,T1);
};

tran *TScale()
{
 tran *T=new tran;
 T->xx=(P2.x-P1.x)/(SC2.x-SC1.x);
 T->yy=(P2.y-P1.y)/(SC2.y-SC1.y);
 T->xo=-T->xx*SC1.x;
 T->yo=-T->yy*SC1.y;
 return(T);
};

tran *TChar(mstep FntSzX,mstep FntSzY)
{
 tran *TC,*TB;
 int  i;
 TB=T;i=SClev;
 while(i--&&(TB!=NULL))TB=TB->next;
 TC=TRot(TB->act.x,TB->act.y,DI.x,DI.y,
  (SI.x*CHsz*CHq)/FntSzX,(SI.y*CHsz)/FntSzY,SL);
 TC->next=TB;
 return(TC);
};

//==============================================================

void ExeDir(char *fn)
{
 char drive[MAXDRIVE];
 char dir[MAXDIR];
 char file[MAXFILE];
 char ext[MAXEXT];
 char file1[MAXFILE];
 char ext1[MAXEXT];
 fnsplit(fn,drive,dir,file,ext);
 fnsplit(ExeName,drive,dir,file1,ext1);
 fnmerge(fn,drive,dir,file,ext);
};

void AddExt(char *fn,char *ext1)
{
 char drive[MAXDRIVE];
 char dir[MAXDIR];
 char file[MAXFILE];
 char ext[MAXEXT];
 if(!(EXTENSION&fnsplit(fn,drive,dir,file,ext)))
  strcpy(fn+strlen(fn),ext1);
};

typedef struct {
    signed   int xoff  : 7;
    unsigned int flag1 : 1;
    signed   int yoff  : 7;
    unsigned int flag2 : 1;
 } BgiSeg;

typedef struct BgiFont{
  HEADER *head;
  word   *off;
  byte   *width;
  byte   *def;
  int    sx,sy,spx;
  int    prop_fl;
  ~BgiFont();
 };

BgiFont::~BgiFont()
{
 free(head);
 free(off);
 free(width);
 free(def);
};

BgiFont *Myfont;

BgiFont *read_font(char name[],int aprop_fl)
{
 FILE *f;
 long l,m;
 AddExt(name,".CHR");
 f=fopen(name,"rb");
 if (f==NULL){
  ExeDir(name);
  f=fopen(name,"rb");
 };
 if (f==NULL) return(NULL);
 BgiFont *b=new BgiFont;
 fseek(f,0x80,SEEK_SET);
 b->head=new HEADER;
 fread(b->head,sizeof(HEADER),1,f);
 b->off=(word *)malloc(256*sizeof(int));
 memset (b->off,0,256*sizeof(int));
 fread(b->off+b->head->first,sizeof(int),b->head->nchrs, f);
 b->width=(byte *)malloc(256*sizeof(byte));
 memset (b->width,0,256*sizeof(byte));
 fread(b->width+b->head->first,b->head->nchrs, sizeof(char), f);
 l = ftell( f);
 fseek( f, 0, SEEK_END );
 m = ftell( f );
 fseek( f, l, SEEK_SET );
 l=m-l;
 b->def=(byte *) malloc( l );
 fread(b->def,l,1, f);
 fclose(f);
 b->sy=b->head->org_to_cap-b->head->org_to_base;
 b->sx=0;
 for(int i=b->head->first,j=i+b->head->nchrs;i<j;i++)
  if(b->sx<b->width[i])b->sx=b->width[i];
 if (b->width['E'])
 {
  int maxx=0,maxy=0,maxmovex=0;
  BgiSeg s,*p=(BgiSeg *)(b->def+b->off['E']);
  s=*p;
  while (s.flag1)
  {
   if(s.flag2){maxx=max(maxx,s.xoff);maxy=max(maxy,s.yoff);}
   else maxmovex=max(maxmovex,s.xoff);
   p++;s=*p;
  };
  b->spx=maxmovex-maxx;
  b->sx=maxx;
  b->sy=maxy;
 };
 b->prop_fl=aprop_fl;
 return(b);
};

void plotchar(PlotterT Plotter,BgiFont *f,byte c)
{
 tran *TC=TChar(f->sx,f->sy);
 point PA;
 BgiSeg s,*p=(BgiSeg *)(f->def+f->off[c]);
 mstep xshft=0;
 if(!f->prop_fl) xshft=(f->sx-f->width[c]+f->spx)/2;
 while (1)
 {
  s=*p;
  if (!s.flag1)
  {
   if(!f->prop_fl)
   {
    DoT(TC,PA,point(f->sx*CHsp.x,0));
    Plotter(round(PA.x),round(PA.y),0);
   };
   DelT(TC);
   return;
  };
  DoT(TC,PA,point(s.xoff+xshft,s.yoff));
  Plotter(round(PA.x),round(PA.y),s.flag2);
  p++;
 };
};

//==============================================================

int DebugLevel=1;
int ToXYfl=0;

PlotterT Plotter=BgiMoveAbs;

#define stdmsk 0xF
word stdend[]={'0'+0x8400,'9'+0x400,'A'+0x8200,'Z'+0x200,
 ';'+0x4000,0x400D,'+'+0x2400,'-'+0x2400,'.'+0x400,0};
word lbend[]={3+0x4000,0+0x8400,0xFF+0x400,0};

int clasify(byte c,word *e)
{
 while (*e){
  byte d=*e&0xFF;
  if(*e&0x8000){if((c>=d)&&(c<=(*(e+1)&0xFF))) return(*e>>8);}
  else if(d==c) return(*e>>8);
  e++;
 };
 return(0);
};

int readfield(FILE *f,byte &fieldtype,byte field[],int maxlen,word *aend)
{
 int i=0;
 static byte ch=0;
 fieldtype=0;field[0]=0;
 byte ft=0;
 if (!ch) if(fread(&ch,1,1,f)!=1) return(0);
 while (!(fieldtype=clasify(ch,aend))) if(fread(&ch,1,1,f)!=1) return(0);
 if (fieldtype&0x2) maxlen=2;  // Nutne pro tisk z Windows !!!!!!
 for(;;){
  field[i++]=ch;field[i]=0;
  if(i==maxlen) {ch=0;return(fieldtype);};
  if(fread(&ch,1,1,f)!=1) {ch=0;return(fieldtype);};
  ft=clasify(ch,aend);
  if ((ft^fieldtype)&stdmsk) return(fieldtype);
 };
};

byte fieldtype=0;
byte field[21]="";
FILE *F=NULL;

int convmstep(mstep &ms,char *field)
{
 int sig=1;long l=0;int i=0;
 if (field[0]=='+') i++;
 if (field[0]=='-') {i++;sig=-1;};
 while (isdigit(field[i])) l=l*10+field[i++]-'0';
 if (field[i]!='.') ms=l*sig;
 else
 {
  i++;int j=MSTEPU;
  while (isdigit(field[i])&&(j>1)) {l=l*10+field[i++]-'0';j=j/10;};
  ms.l=l*j*sig;
 };
 if (field[i]) return(i);
 return(0);
};

int readmstep(mstep &ms)
{
 readfield(F,fieldtype,field,20,stdend);
 if (!(fieldtype&4)) return(-1);
 return(convmstep(ms,field));
};

int pltm=0;

int plot(int sub)
{
 if (sub&1) pltm|=sub;else pltm&=~sub;
 point a;
 for(;;)
 {
  if(readmstep(a.x))return(0);
  if(readmstep(a.y))return(0);
  if (pltm&2) DoT(T,a,a+T->act);
  else DoT(T,a,a);
  Plotter(round(a.x),round(a.y),pltm&4);
 };
};

int SetSI(int sub)
{
 point S;
 if (sub) sub=!readmstep(S.x);
 if (sub) sub=!readmstep(S.y);
 if (!sub){
  if (PS<=3)
   S=point(mstep(285,1000),mstep(375,1000));/*pro A3*/
  else
   S=point(mstep(187,1000),mstep(269,1000));/*pro A4*/
 };
 SI.x=S.x*402;
 SI.y=S.y*402;
 return(0);
};

int SetSR(int sub)
{
 point S;
 if (sub) readmstep(S.x);
 if (sub) readmstep(S.y);
 if (sub){
  SI.x.l=muldiv(labs(P2.x.l-P1.x.l),S.x.l,100l*MSTEPU);
  SI.y.l=muldiv(labs(P2.y.l-P1.y.l),S.y.l,100l*MSTEPU);
 }
 else SetSI(0);
 return(0);
};

int SetDI(int sub)
{
 if (sub) sub=!readmstep(DI.x);
 if (sub) sub=!readmstep(DI.y);
 if (!sub) DI=point(1,0);
 return(0);
};

int SetSL(int sub)
{
 if (sub) sub=!readmstep(SL);
 if (!sub) SL=0;
 return(0);
};

void CharMove(point A,mstep col,mstep lin,int fl)
{
 point B;
 tran *TC=TChar(1,1);
 if(fl) {TC->xo=A.x;TC->yo=A.y;};
 A.x=col*CHsp.x;
 A.y=lin*CHsp.y;
 DoT(TC,B,A);
 Plotter(round(B.x),round(B.y),0);
 DelT(TC);
};

int CPcmd(int sub)
{
 point PA;
 if (sub) sub=!readmstep(PA.x);
 if (sub) sub=!readmstep(PA.y);
 if (!sub) PA=point(0,-1);
 CharMove(PA,PA.x,PA.y,0);
 return(0);
};

int LBcmd(int sub)
{
 int i;
 char ch;
 tran *TB=T;i=SClev;
 while(i--&&(TB!=NULL))TB=TB->next;
 point A=TB->act;
 int Line=0;
 i=2;
 if(sub)
  do
  {
   for(;(ch=field[i])!=0;i++)
   {
    if(ch==0xD) CharMove(A,0,-Line,1);
    else if(ch==0xA) CharMove(A,0,-(++Line),1);
    else plotchar(Plotter,Myfont,ch);
   };
   i=0;
   readfield(F,fieldtype,field,20,lbend);
  }
  while(fieldtype&0x4);
 return(0);
};

int UCcmd(int)
{
 point PA;
 int pen=0;
 tran *TC=TChar(8,8);
 while(1)
 {
  if (readmstep(PA.x)) break;
  if (PA.x.l>=99l*MSTEPU){pen=1;continue;};
  if (PA.x.l<=-99l*MSTEPU){pen=0;continue;};
  if (readmstep(PA.y)) break;
  DoT(TC,PA,TC->act+PA);
  Plotter(round(PA.x),round(PA.y),pen);
 };
 DoT(TC,PA,point(8*CHsp.x,0));
 Plotter(round(PA.x),round(PA.y),0);
 DelT(TC);
 return(0);
};

int SetSC(int sub)
{
 if (SClev&&(sub<2)) {DelT(T);SClev--;};
 if (sub) if((sub=!readmstep(SC1.x))!=0);
 if (sub) if((sub=!readmstep(SC2.x))!=0);
 if (sub) if((sub=!readmstep(SC1.y))!=0);
 if (sub) if((sub=!readmstep(SC2.y))!=0);
 if(DebugLevel>=2)
 {
  printf("Scale ");SC1.x.disp();putch(' ');SC2.x.disp();putch(' ');
  SC1.y.disp();putch(' ');SC2.y.disp();putch(';');
  printf("\n");
 };
 AddT(T,TScale());
 SClev++;
 return(0);
};

void Circle(point C,mstep R,int BA,int EA,int DA)
{
 point A;
 int flg=0;
 for(;DA>0?BA<=EA:BA>=EA;BA+=DA)
 {
  point B(C.x+cosx(R,BA),C.y+sinx(R,BA));
  DoT(T,A,B);
  Plotter(round(A.x),round(A.y),flg);
  flg=1;
 };
};

int CIcmd(int)
{
 mstep R,DA;
 point C,D;
 if (readmstep(R)) return(-1);
 if (readmstep(DA)) DA=5;
 C=T->act;
 Circle(C,R,0,360,round(DA));
 DoT(T,D,C);
 Plotter(round(D.x),round(D.y),0);
 return(0);
};

int ARcmd(int sub)
{
 point C;
 mstep Ams,Dms,R;
 int A,D;
 if (readmstep(C.x)) return(-1);
 if (readmstep(C.y)) return(-1);
 if (readmstep(Ams)) return(-1);
 if (readmstep(Dms)) Dms=1;
 A=round(Ams);
 D=round(Dms);D=labs(D);if(A<0)D=-D;
 if(sub)C=C-T->act;
 R=vectabs(C.x,C.y);
 AddT(T,TRot(T->act.x+C.x,T->act.y+C.y,-C.x,-C.y,1,1));
 Circle(point(0,0),R,0,A,D);
 DelT(T);
 return(0);
};

int SetIP(int sub)
{
 point P=P2-P1;
 if (sub) sub=!readmstep(P1.x);
 if (sub) sub=!readmstep(P1.y);
 if (sub==1) if(readmstep(P2.x)) sub=2;
 if (sub==1) if(readmstep(P2.y)) sub=2;
 if(sub==2) P2=P1+P;
 else if(!sub)
 {
  if (PS<=3)
  {
   P1=point(  561,  306);
   P2=point(15761,10306);
  }else{
   P1=point(  306, 181);
   P2=point(10306,7381);
  };
 };
 TP1P2();
 return(0);
};

int SetDF(int)
{
 SetSI(0);
 SetDI(0);
 SetSL(0);
 SClev=0;
 SetIP(0);
 return(0);
};

int SetPS(int sub)
{
 mstep S;
 if (sub) sub=!readmstep(S);
 if (!sub) S=3;
 PS=round(S);
 SetDF(0);
 return(0);
};

typedef struct{
 int (*fnc)(int sub);
 int sub;
 char*str;
}tsfunc;

tsfunc sfunc[]={
{plot,2,"PA"},
{plot,2+1,"PR"},
{plot,4,"PU"},
{plot,4+1,"PD"},
{SetSC,1,"SC"},
{SetSI,1,"SI"},
{SetSR,1,"SR"},
{SetDI,1,"DI"},
{SetSL,1,"SL"},
{SetPS,1,"PS"},
{SetIP,1,"IP"},
{LBcmd,1,"LB"},
{ARcmd,0,"AR"},
{ARcmd,1,"AA"},
{CIcmd,0,"CI"},
{SetDF,0,"DF"},
{UCcmd,0,"UC"},
{CPcmd,1,"CP"}
};

int readcmd()
{
 int i;
 while(!(fieldtype&2)){
  if (!readfield(F,fieldtype,field,20,stdend))return(0);
 };
 fieldtype=0x80;
 tsfunc *pf=sfunc;
 for (;pf<sfunc+sizeof(sfunc);pf++) if(!strcmpi(pf->str,field))
 {
  pf->fnc(pf->sub);
  return(fieldtype);
 };
 if((field[0]=='L')&&(field[1]=='B')){LBcmd(1);return(fieldtype);};
 if (DebugLevel){printf("Unknown CMD :%s",field);
  while(1)
  {
   if (!readfield(F,fieldtype,field,20,stdend))return(0);
   if (fieldtype&2) break;
   printf(" %s",field);
  };
  printf("\n");
  if (DebugLevel>=2){while (!kbhit());getch();};
 };
 return(fieldtype);
};

int PlotHPGLfile(char *name){
 if ((Myfont=read_font(FontName,Prop_Fl))==NULL){printf("Open font file error \n");exit(1);};
 if (!(F=fopen(name,"rb"))) {printf("Open file error \n");exit(1);};
 setvbuf(F,NULL,_IOFBF, 1024*8);

 pltm=0;
 if(ToXYfl) PenReset();
 SetDF(0);

 while(readcmd())if(kbhit())
 {
  int ch=getch();
  if(ch==0x1B)break;
  if(ch==' '){printf("Paused\n");getch();printf("Continue\n");}
 };
 if(ToXYfl) {MoveAbs(0,0,0);while(!PenRedy());};
 DelAllT();
 return(0);
};

//==============================================================

void BgiMoveAbs(int X,int Y,int Pen)
{
#ifndef __TINY__
 static int Redy=0;
 static int MuX=1,DiX=1,MuY=1,DiY=1,Color=15;
 static int gdriver = DETECT, gmode, errorcode=0;
 if (!Redy)
 {
  initgraph(&gdriver, &gmode, "");
  errorcode=graphresult();
  if (errorcode == grFileNotFound)
  {
   char BgiDir[MAXPATH]="";
   ExeDir(BgiDir);
   gdriver = DETECT;
   initgraph(&gdriver, &gmode,BgiDir);
   errorcode=graphresult();
  };
  if (errorcode != grOk) {printf("Bgi driver error\n");exit(1);};
  MuX = getmaxx();
  MuY = getmaxy();
  DiX = 2640;
  DiY = 1920;
  Redy=1;
 };
 asm{
	mov	ax,X
	imul	MuX
	idiv    DiX
	mov	X,ax
	mov	ax,Y
	imul	MuY
	idiv    DiY
	mov	Y,ax
 };
 setcolor(Color);
 if (Pen)lineto(X,MuY-Y);else moveto(X,MuY-Y);
 if (DebugLevel==5)
  { printf("Pozice : %i %i Pero : %i ",X,Y,Pen);while(!kbhit());getch();};
#endif
};

//==============================================================

void WriteByteXY(int X,int Y,word W)
{
 int I;
 gotoxy(X,Y);for(I=7;I>=0;I--) printf("%1i",I);
 gotoxy(X,++Y);for(I=7;I>=0;I--)
  if (W&(1<<I)) printf("%1i",I); else printf(".");
};

int TestSW(int argc,char *argv[],char swch)
{
 for(int i=1;i<argc;i++)
  if((argv[i][0]=='-')&&(toupper(argv[i][1])==swch)) return(i);
 return(0);
};

void main(int argc,char *argv[])
{
 ExeName=argv[0];
 clrscr();
 int  I;
 int  ch=0;
 if (argc==1)
 {
  printf("\n\
  HPGL interpreter for XY 4141 on LPT1 version "__DATE__"\n\
  (C) PiSOFT 1995\n\
\n\
usage: xy [HPGLfile] [-option] [..] \n\
options:\n\
  p       output to plotter else on screen\n\
  s<n>    scale 0.000 to 999.0000\n\
  f<font> BGI font file defaul LITT.CHR\n\
  t       use proportional text output\n\
  c<n>    relative character size\n\
  q<n>    character x/y ratio\n\
  v<n>    speed from 0 to 6\n\
  3       prevod A3 na format A4\n\
  4       vystup na format A4\n\
  d<n>    debug level (defaul 1)\n\
	   0 no warning\n\
	   1 print unknown commands\n\
	   2 stop and wait on unkown commands\n\
	   10 test mode\n\
\n\
necesary files: EGAVEGA.BGI, LITT.CHR\n\
\n");
  return;
 };
 if(TestSW(argc,argv,'P'))ToXYfl=1;
 if((I=TestSW(argc,argv,'C'))!=0) convmstep(CHsz,argv[I]+2);
 if((I=TestSW(argc,argv,'Q'))!=0) convmstep(CHq,argv[I]+2);
 if((I=TestSW(argc,argv,'S'))!=0) {convmstep(SCu.x,argv[I]+2);SCu.y=SCu.x;};
 if((I=TestSW(argc,argv,'D'))!=0) DebugLevel=atoi(argv[I]+2);
 if((I=TestSW(argc,argv,'F'))!=0) strcpy(FontName,argv[I]+2);
 if((I=TestSW(argc,argv,'T'))!=0) Prop_Fl=1;
 if((I=TestSW(argc,argv,'V'))!=0)
  {int J=atoi(argv[I]+2);if(J>6)J=6; PenSpeed(J,SpdProf4,0);};
 if((I=TestSW(argc,argv,'3'))!=0) PS=3;
 if((I=TestSW(argc,argv,'4'))!=0) PS=4;
 if(ToXYfl) {if (PenInit(0)<0) exit (1); Plotter=MoveAbs;}
 else Plotter=BgiMoveAbs;
 if(TestSW(argc,argv,'N'))Plotter=NulMoveAbs;

 if(DebugLevel<10) {PlotHPGLfile(argv[1]);return;};

 PenInit(0);

 while (ch!=0x1B)
 {
  if (kbhit())
  {
      if (!(ch=getch())) ch=getch()+0x100;
      if ((ch>='0')&&(ch<='9')) OutBuf^=(1<<(ch-'0'));
      switch (ch)
      {
	case 'A','a':PenGen(OutBuf);break;
	case 'E','e':
	 {
	  int X,Y;
	  gotoxy(10,16);printf("                    ");
	  gotoxy(10,16);printf(">");
	  scanf("%i %i",&X,&Y);
	  MoveRel(X,Y);
	  break;
	 };
	case 'F','f':
	 {
	  MoveRel(200,0);
	  MoveRel(0,200);
	  MoveRel(-200,0);
	  MoveRel(0,-200);
	  MoveRel(200,0);
	  MoveRel(0,200);
	  MoveRel(-200,-200);
	  break;
	 };
	case 'G','g':
	 {
	  mstep R=0,A=10;A=A/360;
	  int Fi,X0,Y0;
	  X0=AbsX;Y0=AbsY;
	  PenCtrl(1);
	  for(Fi=0;Fi<=10*360;Fi+=4)
	  {
	   MoveRel(X0-AbsX+round(sinx(A*Fi+R,Fi)),Y0-AbsY+round(cosx(A*Fi+R,Fi)));
	   //printf("\n X=%4i Y=%4i",AbsX,AbsY);
	  }
	  PenCtrl(0);MoveRel(X0-AbsX,Y0-AbsY);break;
	 };
	case 'H','h':
	 {
	  FILE *InFile;
	  if ((InFile=fopen("columbia.grp","rt"))!=NULL)
	  {
	   int  X,Y,PenSt=0;
	   char CH;
	   if (fread(&CH,1,1,InFile)!=1) CH='E';PenCtrl(0);
	   while(CH!='E')
	   {
	    switch (CH)
	    {
	     case 'M':fscanf(InFile,"%i %i",&X,&Y);
		 if(PenSt)PenCtrl(PenSt=0);
		 MoveRel(X/8-AbsX,Y/8-AbsY);
	       break;
	     case 'P':fscanf(InFile,"%i %i",&X,&Y);
		 if(!PenSt)PenCtrl(PenSt=1);
		 MoveRel(X/8-AbsX,Y/8-AbsY);
	       break;
	    };
	    if (fread(&CH,1,1,InFile)!=1) CH='E';
	   };
	   fclose(InFile);
	  };
	  break;
	 };
	case 'I','i':{PenReset();break;};
	case 0x14D:MoveRel(1,0);break;
	case 0x148:MoveRel(0,1);break;
	case 0x14B:MoveRel(-1,0);break;
	case 0x150:MoveRel(0,-1);break;
	case 0x174:MoveRel(100,0);break;
	case 0x18D:MoveRel(0,100);break;
	case 0x173:MoveRel(-100,0);break;
	case 0x191:MoveRel(0,-100);break;
	case 0x152:PenCtrl(1);break;
	case 0x153:PenCtrl(0);break;
	default:gotoxy(10,16);printf("Key =%4X",ch);break;
      };
  };
  WriteByteXY(10,5,OutBuf);
  gotoxy(10,8); printf("Spd= %4i",ActSpd);
  gotoxy(10,9); printf("X =%5i  Y =%5i",ActX,ActY);
  gotoxy(10,10);printf("X =%5i  Y =%5i",AbsX,AbsY);
  outportb(LPPort,OutBuf);
 };
};