/*******************************************************************
  XY Stepper Motor HPGL Plotter

  xy_plt.cpp	- pen and motor optimizing driver

  (C) Copyright 1995,1997 by Pavel Pisa - Original Author
      e-mail:	pisa@cmp.felk.cvut.cz
      homepage:	http://cmp.felk.cvut.cz/~pisa
      work:	http://www.pikron.com/

  This package can be copied and used under one of next licenses

  Gnu General Public License with all its conditions.
  See file GPL for details. Under this license nobody can
  distribute this work and any derived work without full source code
  for all libraries and modules compiled and linked into executable.
  In this case you have right to add your copyright and
  delete next paragraph from this header if you disagree with its
  conditions.
  
  There is second possibility which can be useful for embedded
  and commercial applications. You can use this code as you like
  without any warranties and distribute executables or appliances
  without source. You must send all changes and enhancements made
  in this file and all files included (only) in this module to
  original author (Pavel Pisa) without any further restrictions.

 *******************************************************************/

#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <values.h>

#define StepMsk  0xE
#define StepPuls 0x2
#define StepXN   0x4
#define StepXP   0xC
#define StepYN   0x0
#define StepYP   0x8
#define StepSgn  0x8
#define PenDown  0x1
#define PenUp    0x0

#define TimeIntNo 8

#define I8253_0 0x40;
#define I8253_C 0x43;

typedef unsigned int word;
typedef unsigned char byte;

template <class T>
T max(T x, T y)
{
  return (x > y) ? x : y;
};

word LPPort=0x378;

byte OutBuf;

int AbsMaxSpd=5;
int SpdProf1[]={0,0x8000,0x4000,0x2000,0x1000,0x0C00};
int SpdProf2[]={0,0xA000,0x4000,0x1C00,0x0E00,0x0800};
int SpdProf3[]={0,0x6000,0x2000,0x0C00,0x0600,0x0400};
int SpdProf4[]={0x6000,0x3000,0x2000,0x1400,0x0E00,0x0800,0x0680};
int *SpdProf=SpdProf1;
int PenUpTime=3;

#define SegBufLen 100
typedef struct{
 int  Cmd;
 byte MaxSpd;
 byte EndSpd;
 word Count;
 int  X,Y;
}SegRec;

volatile SegRec SegBuf[SegBufLen];
volatile int InPtr=0,OutPtr=0,ActPtr=0;

int  ActX,ActY,AbsX,AbsY,AbsPen;
byte Dir1,Dir2;
int  D1;
int  D2;
int  Aprox,Count=0;
byte EndSpd;
byte ActSpd;

word ActTick=0;
word NxtTick=0;
word CntTick=0;

int  PenInit(int Port);
void PenReset();
void PenCtrl(int Pen);
void MoveRel(int X,int Y);
void MoveAbs(int X,int Y,int Pen);

void interrupt(*OldTimeInt)(...)=NULL;
int  (*DoFnc)(int Cmd,volatile SegRec * SegAct)=NULL;

int IncSegBuf(int SegPtr)
{
 SegPtr++;
 if (SegPtr==SegBufLen) SegPtr=0;
 return(SegPtr);
};

int DecSegBuf(int SegPtr)
{
 if (SegPtr==0) return(SegBufLen-1);
 return(--SegPtr);
};

int InsSegBuf(SegRec * SegIns)
{
 if (IncSegBuf(InPtr)==OutPtr) return(0);
 SegBuf[InPtr]=*SegIns;
 InPtr=IncSegBuf(InPtr);
 return(1);
};

int OptSegTst(int P)
{
 int I;
 I=SegBuf[P].Cmd;
 I=(I==1)||(I==1);
 return(I);
};

void OptSegBuf(void)
{
 int P,I,OP;
 word S;
 P=InPtr;OP=OutPtr;
 if (P==OP) return;
 P=DecSegBuf(P);
 if ((P==OP)||!OptSegTst(P)) return;
 S=SegBuf[P].EndSpd;
 while (1)
 {
  I=S+SegBuf[P].Count;
  if ((I<S)||(I>AbsMaxSpd)) S=AbsMaxSpd;else S=I;
  P=DecSegBuf(P);

  if (P==OP) return;
  if (!OptSegTst(P)) return;
  if (SegBuf[P].MaxSpd==SegBuf[P].EndSpd) return;
  if (S==SegBuf[P].EndSpd) return;
  if (S>SegBuf[P].MaxSpd) S=SegBuf[P].MaxSpd;
  SegBuf[P].EndSpd=S;
 };
};

void SetTick(word Tick)
{
 asm{
	pushf
	cli
	mov	dx,I8253_C
	mov	al,0x34
	out     dx,al
	mov	dx,I8253_0
	mov	ax,[Tick]
	mov     [NxtTick],ax
	out     dx,al
	mov     al,ah
	out     dx,al
	popf
 };
};

void PenGen(byte Dir)
{
 asm{
	mov	dx,[LPPort]
	mov     al,[OutBuf]
	and     al,not StepMsk
	or      al,[Dir]
	or      al,StepPuls
	xor     al,StepPuls
	mov     cx,10
 };a1:asm{
	out     dx,al
	loop    a1
	xor     al,StepPuls
	mov     cx,10
 };a2:asm{
	out     dx,al
	loop    a2
	mov     [OutBuf],al
 };
 switch (Dir)
 {
  case StepXN:ActX--;break;
  case StepXP:ActX++;break;
  case StepYN:ActY--;break;
  case StepYP:ActY++;break;
 };
};

void AdjSpd(void)
{
 int C;
 C=ActSpd-Count+1-EndSpd;
 if ((C>0)&&ActSpd) ActSpd--;
 else if ((C<0)&&(ActSpd<AbsMaxSpd))ActSpd++;
};

int DoMoveRel(int Cmd,volatile SegRec *SegAct)
{
 if (Cmd)
 {
  D1=SegAct->X;
  D2=SegAct->Y;
  EndSpd=SegAct->EndSpd;
  if (abs(D1)>=abs(D2)) {Dir1=StepXP;Dir2=StepYP;}
   else {Dir1=StepYP;Dir2=StepXP;Aprox=D1;D1=D2;D2=Aprox;};
  if (D1<0) {D1=-D1;Dir1^=StepSgn;};
  if (D2<0) {D2=-D2;Dir2^=StepSgn;};
  Count=D1;Aprox=-D1/2;
 }
 else
 {
  Count--;
  if((Aprox+=D2)>0){Aprox-=D1;PenGen(Dir2);};
  PenGen(Dir1);
 };
 AdjSpd();
 return(Count);
};

void MoveRel(int X,int Y)
{
 int P;
 SegRec Seg1={1,0,0,0,0};
 Seg1.X=X;AbsX+=X;
 Seg1.Y=Y;AbsY+=Y;
 if(!(Seg1.Count=max(abs(X),abs(Y))))return;
 if ((InPtr!=OutPtr)&&(SegBuf[P=DecSegBuf(InPtr)].Cmd==1))
 {
  long L1,L2;
  int  A1,A2;
  A1=max(abs(X),abs(Y));
  A2=max(abs(SegBuf[P].X),abs(SegBuf[P].Y));
  L1=max(labs((long)X*A2-(long)SegBuf[P].X*A1),
   labs((long)Y*A2-(long)SegBuf[P].Y*A1));
  L2=(long)A1*A2;
  if (L1<(MAXLONG/AbsMaxSpd)) A1=AbsMaxSpd-L1*AbsMaxSpd/L2;
  else A1=AbsMaxSpd-L1/(L2/AbsMaxSpd);
  if (A1>0) SegBuf[P].MaxSpd=A1;
 };
 while (!InsSegBuf(&Seg1));
 OptSegBuf();
};

void MoveAbs(int X,int Y,int Pen)
{
 if (!AbsPen!=!Pen) PenCtrl(Pen);
 MoveRel(X-AbsX,Y-AbsY);
};

int DoPenCtrl(int Cmd,volatile SegRec *SegAct)
{
 if(Cmd)
 {
  Count=SegAct->Count;
  OutBuf&=~(PenUp|PenDown);
  if(SegAct->X) OutBuf|=PenDown; else OutBuf|=PenUp;
  outportb(LPPort,OutBuf);
  ActSpd=0;
 }
 else
 {
  Count--;
 };
 return(Count);
};

void PenCtrl(int Pen)
{
 AbsPen=Pen;
 SegRec Seg1={2,0,0,0,0};
 Seg1.X=Pen;
 Seg1.Count=PenUpTime;
 while (!InsSegBuf(&Seg1));
 OptSegBuf();
};

void PenReset(){
 PenCtrl(0);MoveRel(0,-2500);ActY=2500;
 ActX=AbsX=AbsY=0;
};

void interrupt TimeInt(...)
{
 int T=0;
 if (!(OutBuf&0x80))
 {
  if (DoFnc!=NULL) T=DoFnc(0,&(SegBuf[ActPtr]));
  if (!T)
  {
   DoFnc=NULL;
   if(OutPtr!=InPtr)
   {
    ActPtr=OutPtr;OutPtr=IncSegBuf(OutPtr);
    switch (SegBuf[ActPtr].Cmd)
    {
     case 1:DoFnc=DoMoveRel;break;
     case 2:DoFnc=DoPenCtrl;break;
    };
    DoFnc(1,&(SegBuf[ActPtr]));
   };
  };
 };
 poke(0xB800,0,ActSpd|0x7030);
 CntTick+=ActTick;
 if (CntTick-ActTick<CntTick)
  { outp(0x20,0x20);ActTick=NxtTick;SetTick(SpdProf[ActSpd]);return;}
 SetTick(SpdProf[ActSpd]);
 ActTick=NxtTick;
 OldTimeInt();
};

int PenRedy()
{
 return((InPtr==OutPtr)&&(DoFnc==NULL));
};

void Ending()
{
 disable();
 if (OldTimeInt!=NULL) setvect(TimeIntNo,OldTimeInt);
 OldTimeInt=NULL;
 SetTick(0);
 enable();
};

void Abort_program(int Sig)
{
  printf("!! User BREAK detected !! Type : %d\n",Sig);
  exit(1);
};

int PenInit(word Port)
{
 if (OldTimeInt!=NULL) return(-2);
 OutBuf=0;
 if (Port) LPPort=Port;
 if (atexit(Ending)) return (-1);
 signal(SIGABRT,Abort_program);
 signal(SIGINT,Abort_program);
 disable();
 OldTimeInt=getvect(TimeIntNo);
#ifdef __TINY__
 setvect(TimeIntNo,(void interrupt(*)(...))MK_FP(_CS,TimeInt));
#else
 setvect(TimeIntNo,TimeInt);
#endif
 enable();
 return(0);
};

void PenSpeed(int AAbsMaxSpd,int *ASpdProf,int APenUpTime)
{
 if(APenUpTime) PenUpTime=APenUpTime;
 if(ASpdProf!=NULL) SpdProf=ASpdProf;
 if(AAbsMaxSpd) AbsMaxSpd=AAbsMaxSpd;
};